from .main import main_blueprint
from .user import user_blueprint
from .game import game_blueprint
from .watcher import watcher_blueprint


def register_blueprints(app):
    app.register_blueprint(main_blueprint)
    app.register_blueprint(user_blueprint, url_prefix='/user')
    app.register_blueprint(game_blueprint, url_prefix='/game')
    app.register_blueprint(watcher_blueprint, url_prefix='/watcher')



from flask_user import UserMixin
from sqlalchemy import DateTime
from datetime import datetime
from enum import IntEnum, unique, auto

from lamon import db


class User(db.Model, UserMixin):
    """ A User

    :param username: Username to authenticate with lamon
    :param password: Hashed password
    :param nicknames: :class:`Nickname` s of the User
    :param events: Events concerning the user (like scores, joins, etc.)
    :param roles: System roles of the user
    """
    __tablename__ = 'users'

    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(50), unique=True)
    password = db.Column(db.String(255))
    is_enabled = True

    nicknames = db.relationship('Nickname', back_populates='user')
    events = db.relationship('Event', back_populates='user')
    roles = db.relationship('Role', secondary='user_roles',
                            backref=db.backref('users', lazy='dynamic'))

    def __repr__(self):
        return self.username


class Event(db.Model):
    """ Events are used log keep a log of things happening. This includes events
    on watched servers (scores, kills), server events (unavailable) and
    watcher events (restarted, started, stopped).

    :param type: An :class:`EventType` specifies how to handle this event
    :param time: Time the event occurred
    :param info: Additional info. Changes with :attr:`type`
    :param watcher: Watcher the event occurred in (If any)
    :param game: Game the event occured in (If any)
    :param user: User connected to the event (If any)
    """
    __tablename__ = 'events'

    id = db.Column(db.Integer, primary_key=True)
    type = db.Column(db.Integer)
    time = db.Column(DateTime())
    info = db.Column(db.String)

    watcherID = db.Column(db.Integer, db.ForeignKey('watchers.id'))
    watcher = db.relationship('Watcher', back_populates='events')

    gameID = db.Column(db.Integer, db.ForeignKey('games.id'))
    game = db.relationship('Game', back_populates='events')

    userID = db.Column(db.Integer, db.ForeignKey('users.id'))
    user = db.relationship('User', back_populates='events')

    def __init__(self, type=None, **kwargs):
        super().__init__(type=int(type), **kwargs)

    def __str__(self):
        msg = str(EventType(self.type)) + ": "

        if self.watcherID is not None:
            msg += "Watcher = {}; ".format(str(self.watcher))

        if self.gameID is not None:
            msg += "Game = {}; ".format(str(self.game))

        if self.userID is not None:
            msg += "Gmae = {}; ".format(str(self.game))

        return msg


class Game(db.Model):
    """ A game or gamemode played by the users

    :param name: Name of the game, can be anything
    :param nicknames: Nicknames registered for this game
    :param events: Events in the Game
    :param watchers: Watchers watching the game
    """
    __tablename__ = 'games'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String, unique=True)
    nicknames = db.relationship('Nickname', back_populates='game')
    events = db.relationship('Event', back_populates='game')
    watchers = db.relationship('Watcher', back_populates='game')

    def __repr__(self):
        return self.name


class Watcher(db.Model):
    """ A Watcher is a process keeping track of one or more gameservers

    :param threadClass: Implementation of the process
    :param config: Configuration values the process needs (like server ip)
    :param events: Events occured in the watcher
    :param game: Game watched by the watcher
    """
    __tablename__ = 'watchers'

    id = db.Column(db.Integer, primary_key=True)
    threadClass = db.Column(db.String)
    config = db.relationship('WatcherConfig', back_populates='watcher')

    events = db.relationship('Event', back_populates='watcher')

    gameID = db.Column(db.Integer, db.ForeignKey('games.id'))
    game = db.relationship('Game', back_populates='watchers')

    def __repr__(self): return "{} ({})".format(
        self.game, self.config)


class Role(db.Model):
    """ A user role on the system

    :param name: Name of the role
    """
    __tablename__ = 'roles'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String, unique=True)

    def __repr__(self):
        return self.name


class UserRoles(db.Model):
    """ Many-to-many mapper between :class:`User` and :class:`Role` """
    __tabelname__ = 'user_roles'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer(), db.ForeignKey(
        'users.id', ondelete='CASCADE'))
    role_id = db.Column(db.Integer(), db.ForeignKey(
        'roles.id', ondelete='CASCADE'))


class Nickname(db.Model):
    """ The nickname of a user in a game

    :param nick: Nickname
    :param user: Associated user
    :param game: Associated game
    """
    __tablename__ = 'nicknames'

    id = db.Column(db.Integer, primary_key=True)
    nick = db.Column(db.String)

    userID = db.Column(db.Integer, db.ForeignKey(
        'users.id', ondelete='CASCADE'))
    user = db.relationship('User', back_populates='nicknames')

    gameID = db.Column(db.Integer, db.ForeignKey(
        'games.id', ondelete='CASCADE'))
    game = db.relationship('Game', back_populates='nicknames')

    def __repr__(self):
        r = "{} is {} in {}"
        return r.format(self.user, self.nick, self.game)


class WatcherConfig(db.Model):
    """ A key value pair in a :class:`Watcher` configuration

    :param watcher: Associated watcher
    :param key: Key
    :param value: Value
    """
    __tablename__ = 'watcherconfig'

    id = db.Column(db.Integer, primary_key=True)

    watcherID = db.Column(db.Integer, db.ForeignKey('watchers.id'))
    watcher = db.relationship('Watcher', back_populates='config')

    key = db.Column(db.String)
    value = db.Column(db.String)

    def __repr__(self):
        return self.key + ": " + self.value


@unique
class EventType(IntEnum):
    """ A Type of event used in :attr:`~Event.type`

    **Watcher related events**. :attr:`Event.watcherID` has to be set

    :param WATCHER_RELOAD: Watcher reloaded it's config.
    :param WATCHER_START: Watcher started
    :param WATCHER_STOP: Watcher stopped
    :param WATCHER_CONNECTION_LOST: Watcher lost connection
    :param WATCHER_CONNECTION_REAQUIRED: Watcher got connection back

    **User related events**. :attr:`Event.userID`, :attr:`Event.watcherID` and
    :attr:`Event.gameID` have to be set.

    :param USER_SCORE: User score. :attr:`Event.info` is the absolute score.
    :param USER_JOIN: User joined a game.
    :param USER_LEAVE: User left a game.
    :param USER_DIE: User died in game
    :param USER_RESPAWN: User respawned after death
    """
    WATCHER_RELOAD = 1000
    WATCHER_START = 1001
    WATCHER_STOP = 1002
    WATCHER_CONNECTION_LOST = 1003
    WATCHER_CONNECTION_REAQUIRED = 1004

    USER_SCORE = 2000
    USER_JOIN = 2001
    USER_LEAVE = 2002
    USER_DIE = 2003
    USER_RESPAWN = 2004



import logging

from flask import current_app
from sqlalchemy.orm import scoped_session, sessionmaker

from ..models import Watcher as WatcherModel
from ..watcher import Watcher, load_watcher_class


class WatcherManager():
    """ Manage watchers in a Flask application.

    :param app: Flask app to bind to
    :param db: Flask-SQLAlchemy Database to use
    """

    def __init__(self, app=None, db=None):
        self.app = app
        self.db = db
        self._watchers = {}

        if app:
            self.init_app(app, db)
        else:
            self.logger = logging.getLogger('lamon.watcher_manager')

    def init_app(self, app, db):
        """ Attach to application

        :param app: Flask app to attach to
        :param db: Flask-SQLAlchemy Database to use
        """
        app.watcher_manager = self
        self.db = db
        self.logger = app.logger.getChild('watcher_manager')

    def start(self, id=None, model=None):
        """ Start a watcher.

        Either id or model is required as argument

        :type id: int
        :param id: Database id of watcher (Default value = None)

        :type model: :class:`lamon.model.Watcher`
        :param model: Database model of watcher (Default value = None)

        :raises ValueError: When watcher is already running
        :raises TypeError: When threadClass of watcher model is no subclass of
            Watcher
        """
        if id is not None:
            model = WatcherModel.query.filter(WatcherModel.id == id).one()
        elif model is not None:
            id = model.id

        if id in self._watchers:
            self.logger.warning(
                "Cannot start already running watcher: {}".format(model))
            raise ValueError(
                "Cannot start watcher (id={}). Already running.".format(id))

        self.logger.info("Starting gamewatcher: {}".format(model))

        # Create db session not connected to flask
        session = scoped_session(sessionmaker(bind=self.db.engine))

        watcher_ = load_watcher_class(model.threadClass)
        watcher = watcher_(session=session, model_id=id)

        self._watchers[id] = watcher
        watcher.start()

    def reload(self, id=None, model=None):
        """ Reload the configuration of a watcher

        Either id or model is required as argument

        :type id: int
        :param id: Database id of watcher (Default value = None)

        :type model: :class:`lamon.model.Watcher`
        :param model: Database model of watcher (Default value = None)

        :raises ValueError: When watcher is not running
        """
        if id is not None:
            model = WatcherModel.query.filter(WatcherModel.id == id).one()
        elif model is not None:
            id = model.id

        if id in self._watchers:
            self._watchers[id].reload()
        else:
            self.logger.warning("Cannot reload watcher: {}".format(model))
            raise ValueError(
                "Cannot reload watcher (id={}). Not running".format(id))

    def stop(self, id=None, model=None):
        """ Stop a watcher

        Either id or model is required as argument

        :type id: int
        :param id: Database id of watcher (Default value = None)

        :type model: :class:`lamon.model.Watcher`
        :param model: Database model of watcher (Default value = None)

        :raises ValueEror: When watcher is not running
        """
        if id is not None:
            model = WatcherModel.query.filter(WatcherModel.id == id).one()
        elif model is not None:
            id = model.id

        if id in self._watchers:
            self.logger.info("Stopping watcher: {}".format(model))
            self._watchers[id].stop()
            self._watchers.pop(id)
        else:
            self.logger.warning("Watcher thread not found: {}".format(model))
            raise ValueError(
                "Cannot stop watcher (id={}). Not running".format(id))

    def is_running(self, id=None, model=None):
        """ Check if given watcher is currently running

        Either id or model is required as argument

        :type id: int
        :param id: Database id of watcher (Default value = None)

        :type model: :class:`lamon.model.Watcher`
        :param model: Database model of watcher (Default value = None)

        :returns: :class:`bool`
        """
        if id is None:
            id = model.id

        return id in self._watchers



from flask import request, redirect
from flask_admin import Admin, expose
from flask_admin.contrib.sqla import ModelView
from flask_user import current_user, login_required, roles_required

from .models import User, Role, Game, Watcher, Nickname, WatcherConfig
from . import db


class AuthModelView(ModelView):
    @login_required
    def is_accessible(self):
        return current_user.has_roles('admin')


class UserView(AuthModelView):
    can_view_details = True

    column_list = ('username', 'roles', 'nicknames')
    inline_models = [Nickname, Role]


class GameView(AuthModelView):
    column_list = ('name', 'nicknames', 'watchers')
    form_excluded_columns = ['scores', 'watchers']
    inline_models = [Nickname]


class WatcherView(AuthModelView):
    column_list = ('threadClass', 'game', 'user', 'config')
    inline_models = [WatcherConfig, ]
    form_choices = {
        'threadClass': [
            ('lamon.watcher.plugins.source_engine.SourceEngineWatcher',
                'Source Engine Watcher'),
            ('lamon.watcher.plugins.quake3.Quake3Watcher',
                'Quake3 Watcher'),
            ('tests.test_watcher.fake.FakeWatcher',
                'Fake Watcher')
        ]
    }

    def get_save_return_url(self, model, is_created=False):
        """ Return the user to edit page after watcher creation. """
        if is_created is True:
            if request.path == self.get_url('.create_view'):
                return self.get_url('.edit_view', id=model.id)+'&created=true'

        return self.get_url('.index_view')

    def create_form(self, obj=None):
        """ The create form should only select the threadClass of the watcher """
        form = super().create_form(obj)
        del form.game
        del form.config
        return form

    def edit_form(self, obj=None):
        form = super().edit_form(obj)

        if request.args.get('created', 'false') == 'true':
            del form.threadClass

        return form


def register_admin(app):
    admin = Admin(app, name='lamon', template_mode='bootstrap3')

    admin.add_view(UserView(User, db.session))
    admin.add_view(GameView(Game, db.session))
    admin.add_view(WatcherView(Watcher, db.session))



import logging
import logging.config
import time
import toml

from flask import Flask, request
from flask.logging import default_handler
from flask_sqlalchemy import SQLAlchemy
from flask_user import UserManager, SQLAlchemyAdapter
from sqlalchemy import event
from sqlalchemy.engine import Engine

__version__ = '0.1.0'

db = SQLAlchemy()


def dictDefault(d, default):
    """ Load default values into a dict """
    if isinstance(d, dict):
        for key, value in default.items():
            if key not in d:
                d[key] = value
            else:
                if isinstance(d[key], dict) and isinstance(default[key], dict):
                    d[key] = dictDefault(d[key], default[key])

    return d


def load_config_file(config_file):
    """ Load a configuration file """
    default_config = {
        'logging': {
            'version': 1,
            'root': {
                'level': 'INFO',
                'handlers': ['console']
            },
            'loggers': {
                'lamon.db': {
                    'level': 'WARNING'
                },
                'flask.app.watcher_manager': {
                    'level': 'WARNING'
                },
                'flask.app.requests': {
                    'level': 'WARNING'
                }
            },
            'formatters': {
                'default': {
                    'format': '%(name)s - %(levelname)s - %(message)s'
                }
            },
            'handlers': {
                'console': {
                    'class': 'logging.StreamHandler',
                    'stream': 'ext://sys.stdout',
                    'formatter': 'default'
                }
            }
        }
    }

    config = dictDefault(toml.load(config_file), default_config)

    config['flask'] = {}
    config['flask']['SECRET_KEY'] = config['database']['secret_key']
    config['flask']['SQLALCHEMY_DATABASE_URI'] = config['database']['database_uri']

    for key, value in config['app'].items():
        config['flask'][key.upper()] = value

    return config


def init_db(app):
    from .models import User, Role, UserRoles

    with app.app_context() as ctx:
        role = None

        if Role.query.filter(Role.name == 'admin').count() == 0:
            app.logger.debug('Creating default role "admin"')
            role = Role(name='admin')
            db.session.add(role)
            db.session.commit()
        else:
            role = Role.query.filter(Role.name == 'admin').one()

        if User.query.filter(User.username == 'admin').count() == 0:
            app.logger.debug('Creating default user "admin" (password: 1234)')
            password = app.user_manager.hash_password('1234')
            user = User(username='admin', password=password, roles=[role])
            db.session.add(user)
            db.session.commit()


def create_app(config_file='config.toml'):
    # Load config
    config = load_config_file(config_file)

    # Configure logging
    logging.config.dictConfig(config['logging'])

    # Create Flask
    app = Flask(__name__)
    app.config.from_mapping(config['flask'])
    app.config['USER_ENABLE_EMAIL'] = False
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

    # Request logging
    @app.after_request
    def log_request(response):
        logger = app.logger.getChild('requests')
        logger.debug("{} - {} - {}".format(request.method,
                                           response.status_code, request.url))
        return response

    # DB logging
    db_logger = logging.getLogger("lamon.db")

    if db_logger.level <= logging.INFO:
        @event.listens_for(Engine, "before_cursor_execute")
        def before_cursor_execute(conn, cursor, statement,
                                    parameters, context, executemany):
            conn.info.setdefault('query_start_time', []).append(time.time())

        @event.listens_for(Engine, "after_cursor_execute")
        def after_cursor_execute(conn, cursor, statement,
                                    parameters, context, executemany):
            total = time.time() - conn.info['query_start_time'].pop(-1)

            if total > 0.01:
                deb = db_logger.info
            else:
                deb = db_logger.debug

            deb("Query Complete: %s", statement)
            deb("Total Time: %f", total)

    # Database
    from .models import User, Event, Game, Watcher, Role, UserRoles, Nickname, \
        WatcherConfig
    db.init_app(app)
    db.create_all(app=app)

    # User Manager
    db_adapter = SQLAlchemyAdapter(db, User)
    user_manager = UserManager(db_adapter, app)
    init_db(app)

    # Admin views
    from .admin import register_admin
    register_admin(app)

    # Register blueprints
    from .views import register_blueprints
    register_blueprints(app)

    # Start Watchers
    from .watcher.manager import WatcherManager
    watcher_manager = WatcherManager(app, db)

    return app



from flask import Blueprint, render_template
from flask_user import login_required

user_blueprint = Blueprint('users', __name__, template_folder='templates')

@user_blueprint.route('/')
def index():
    return render_template('user/index.html')


@user_blueprint.route('/stats')
@user_blueprint.route('/stats/<string:username>')
@login_required
def stats(username=None):
    return render_template('user/stats.html')



from flask import Blueprint, render_template, request, current_app
from ..models import Game


game_blueprint = Blueprint('games', __name__, template_folder='templates')


@game_blueprint.route('/')
def index():
    return render_template('game/index.html')


@game_blueprint.route('/stats/')
@game_blueprint.route('/stats/<int:game_id>')
def stats(game_id=None):
    if game_id is None:
        games = Game.query.all()
        return render_template('game/stats.html', games=games)
    else:
        game = Game.query.filter(Game.id == game_id).one()
        return render_template('game/stats_one.html', game=game)



from flask import Blueprint, render_template, request, current_app, flash, redirect, url_for
from flask_user import roles_required, current_user
from sqlalchemy.orm.exc import NoResultFound

from .. import db
from ..forms import WatcherControlForm, WatcherCreateForm, WatcherEditForm
from ..models import Watcher, Game, WatcherConfig
from ..watcher import load_watcher_class

watcher_blueprint = Blueprint(
    'watchers', __name__, template_folder='templates')


@watcher_blueprint.route('/')
@watcher_blueprint.route('/<int:watcher_id>')
def index(watcher_id=None):
    if watcher_id:
        watcher = Watcher.query.filter(Watcher.id == watcher_id).one()
        form = WatcherControlForm.for_watcher(watcher.id)()

        return render_template('watcher/index_one.html',
                               watcher=watcher, form=form)
    else:
        watchers = []
        show_controls = current_user.is_authenticated and current_user.has_role(
            'admin')

        for watcher in db.session.query(Watcher).all():
            if show_controls:
                form = WatcherControlForm.for_watcher(watcher.id)()
            else:
                form = None

            watchers.append((watcher, form))

        return render_template('watcher/index.html', watchers=watchers,
                               show_controls=show_controls)


@watcher_blueprint.route('/<int:watcher_id>/edit', methods=['GET', 'POST'])
@roles_required(['admin'])
def edit(watcher_id):
    watcher_model = Watcher.query.filter(Watcher.id == watcher_id).one()
    watcher_class = load_watcher_class(watcher_model.threadClass)

    form = WatcherEditForm.load_config(watcher_class, watcher_model)()
    form.game.query = Game.query

    if request.method == 'POST':
        if form.validate_on_submit():
            watcher_model.game = form.data['game']

            # XXX This is probably ugly
            for key, opts in watcher_class.config_keys.items():
                try:
                    watcher_config = WatcherConfig.query.\
                        filter(WatcherConfig.watcher == watcher_model).\
                        filter(WatcherConfig.key == key).\
                        update({'value': form.data[key]})
                except NoResultFound:
                    watcher_config = WatcherConfig(watcher=watcher_model,
                                                   key=key,
                                                   value=form.data[key])
                    db.session.add(watcher_config)

            db.session.commit()

            return redirect(url_for('watchers.index', watcher_id=watcher_id))
        else:
            flash(form.errors, 'warning')

    return render_template('watcher/edit.html', watcher=watcher_model, form=form)


@watcher_blueprint.route('/<int:watcher_id>/stats')
def stats(watcher_id):
    watcher = Watcher.query.filter(Watcher.id == watcher_id).one()
    return render_template('watcher/stats.html', watcher=watcher)


@watcher_blueprint.route('/control', methods=['POST'])
@roles_required(['admin'])
def control():
    app = current_app
    form = WatcherControlForm()

    if form.validate_on_submit():
        try:
            if form.data['start']:
                app.watcher_manager.start(id=form.data['watcher_id'])
                flash('Started watcher', 'success')
            elif form.data['stop']:
                app.watcher_manager.stop(id=form.data['watcher_id'])
                flash('Stopped watcher', 'success')
            elif form.data['reload']:
                app.watcher_manager.reload(id=form.data['watcher_id'])
                flash('Reloaded watcher', 'success')
        except ValueError as e:
            flash(str(e), 'warning')
    else:
        flash(form.errors, 'warning')

    return redirect(request.referrer or url_for('watchers.index'))


@watcher_blueprint.route('/create', methods=['GET', 'POST'])
@roles_required(['admin'])
def create():
    app = current_app

    form = WatcherCreateForm()

    if request.method == 'POST':
        if form.validate_on_submit():
            watcher = Watcher(threadClass=form.data['watcher_class'])
            db.session.add(watcher)
            db.session.commit()
            return redirect(url_for('watchers.edit', watcher_id=watcher.id))

    return render_template('watcher/create.html', form=form)



import sys

from abc import ABC, abstractmethod
from threading import Thread
from importlib import import_module
from logging import getLogger, StreamHandler
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from datetime import datetime

from lamon import db
from ..models import Watcher as WatcherModel
from ..models import Nickname, Event, WatcherConfig, EventType, User


class Watcher(ABC, Thread):
    """ Abstract watcher class. Has to be extended to create a "real"
    watcher plugin.

    :type logName: :class:`str`
    :param logName: Name of the logger used by the watcher.
        When extending using __name__ is a good idea

    :type session: :class:`sqlalchemy.orm.session.Session`
    :param session: SQLAlchemy session to access the database. You probably
        want a scoped_session

    :type model_id: :class:`int`
    :param model_id: The database id (primary key) of the model.

    :raises TypeError: When initialized with a model whose threadClass does
        not fit the object
    """

    config_keys = {}
    """ Plugins can specify which configuration values the user should be able
    to set with this dict.

    Each key of the dict is the name of a configuration key in the database.
    The value is a dict with the following keys:

        * *type*: Function to typecast a string (from the database) into
            a usable value (Required)
        * *required*: Whether this value is required (Required)
        * *hint*: Text to show when editing this configuration value (Optional)
        * *warning*: Warning to show when editing this configuration value (Optional)

    Example:

    .. code-block:: python

        config_keys = {'address': {'type': 'str', 'required': True},
                    'port': {'type': int, 'required': True},
                    'timeout': {'type': int, 'required': True},
                    'app_id:': {'type': int, 'required': True,
                                'hint':
                                \"\"\"this is the app ID of the client, not the
                                server. For example, for Team Fortress 2 440 has
                                to be used instead of 232250 which is the ID of
                                the server software.\"\"\"}}
    """

    def __init__(self, logName, session=None, model_id=None):
        self._session = session
        self._model_id = model_id
        self.config = {}

        # Setup logger
        self.logger = getLogger(logName)

        # Check for correct model
        qualname = type(self).__module__ + "." + type(self).__name__
        if qualname != self._model.threadClass:
            raise TypeError("""Cannot initalize watcher of class {} with model
                            where threadClass == {}""".
                            format(qualname, self._model.threadClass))

        self.logger.propagate = False
        self.reload()
        self.logger.propagate = True

        # Initialize threading.Thread
        super().__init__(name='Watcher-{}'.format(self._model.id))

    def run(self):
        """ Runs Watcher.runner in a new thread."""
        # Log start
        self.logger.info('Started watcher (id={})'.format(self._model_id))
        self.add_event(Event(type=EventType.WATCHER_START))

        try:
            self.runner()
        except Exception as e:
            self.logger.exception(e)

        self._session.commit()
        self._session.close()

    @abstractmethod
    def runner(self):
        """ Main loop. Has to be overwritten.

        This function is supposed to run until Watcher.shutdown is set to true
        or an error occurs. Restarting and error logging is handled by
        Watcher.run

        Errors indicating a connection problem should not crash this function.
        """

    @property
    def _model(self):
        """ Since some database systems don't like multithreaded access
        this reloads the model at every access.
        """
        query = self._session.query(WatcherModel).\
            filter(WatcherModel.id == self._model_id)
        return query.one()

    def reload(self):
        """ Reload all config keys (specified in Watcher.config_keys)
        """
        for key, value in self.config_keys.items():
            query = self._session.query(WatcherConfig).\
                filter(WatcherConfig.watcherID == self._model_id).\
                filter(WatcherConfig.key == key)
            try:
                self.config[key] = value['type'](query.one().value)
                self.logger.debug(
                    "Reload: {} = {}".format(key, self.config[key]))
            except NoResultFound:
                if value['required']:
                    self.logger.warning(
                        "No config w/ key found: {}".format(key))
                else:
                    self.logger.debug("Optional key {} not found".format(key))

        self.logger.info("Reloaded Watcher (id={})".format(self._model.id))
        self.add_event(Event(type=EventType.WATCHER_RELOAD))

    def add_score(self, nickname, score):
        """ Add a score to the user.

        This is a convenience function over :meth:`add_event` and
        :meth:`get_user`

        :type nickname: str
        :param nickname: Nickname of the user

        :type score: int or float
        :param score: **Absolute** score of the user

        :raises ValueError: When no user with the given nickname is found
        """
        user = self.get_user(nickname)
        self.add_event(Event(userID=user.id, gameID=self._model.gameID,
                             type=EventType.USER_SCORE, info=str(score)))

    def add_event(self, event):
        """ Add a event to the database

        :type event: lamon.models.Event
        :param event: Event to add to the database. `event.watcherID` and
            `event.time` are set by this function
        """
        event.watcherID = self._model_id

        if event.time is None:
            event.time = datetime.now()

        self._session.add(event)
        self._session.commit()

        self.logger.debug(str(event))

    def get_user(self, nickname):
        """ Get the :class:`lamon.models.User` with the associated Nickname

        :type nickname: str
        :param nickname: Nickname

        :raises ValueError: When no user with the given nickname is found
        """
        query = self._session.query(User).join(Nickname).\
            filter(Nickname.nick == nickname).\
            filter(Nickname.gameID == self._model.game.id)

        try:
            return query.one()
        except NoResultFound:
            raise ValueError("No user with given nickname ({}) found".
                             format(nickname))

    def stop(self):
        """ Stops the watcher """
        self.logger.info("Stopping watcher")
        self.shutdown = False

        self.join()
        self.logger.debug("Watcher thread stopped")

        self.add_event(Event(type=EventType.WATCHER_STOP))

        self._session.commit()
        self._session.close()
        self.logger.debug("DB session closed")

    def __repr__(self):
        return str(self._model)


class WatcherException(Exception):
    """ Exception to be raised when a watcher encounters problems. (Provided
    there is no better Exception)
    """
    pass


def load_watcher_class(className):
    """ Dynamically load a class extending watcher into the module namespace.
    This is used by WatcherManager to load the class at watcher startup.

    :type className: str
    :param className: Name of the class (with module) to be loaded

    :raises TypeError: When the provided class does not extend :class:`Watcher`

    :returns: Subclass of :class:`Watcher`
    """
    module = import_module(".".join(className.split(".")[:-1]))

    watcher_ = getattr(module, className.split(".")[-1])

    if not issubclass(watcher_, Watcher):
        raise TypeError(
            "{} is not a subclass of watcher".format(str(watcher_)))

    return watcher_



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['lamon', 'lamon.views', 'lamon.watcher', 'lamon.watcher.plugin']

package_data = \
{'': ['*'],
 'lamon': ['static/bootstrap/css/*',
           'static/bootstrap/js/*',
           'static/css/*',
           'templates/*',
           'templates/game/*',
           'templates/main/*',
           'templates/user/*',
           'templates/watcher/*']}

install_requires = \
['flask-admin>=1.5,<2.0',
 'flask-sqlalchemy>=2.3,<3.0',
 'flask-user==0.6.21',
 'flask-wtf>=0.14.2,<0.15.0',
 'flask>=1.0,<2.0',
 'jinja2>=2.10,<3.0',
 'python-valve>=0.2.1,<0.3.0',
 'sqlalchemy>=1.2,<2.0',
 'toml>=0.10.0,<0.11.0']

setup_kwargs = {
    'name': 'lamon',
    'version': '0.1.0',
    'description': 'Monitor gameservers',
    'long_description': None,
    'author': 'Fabian Geiselhart',
    'author_email': 'me@f4814n.de',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from flask_wtf import FlaskForm
from wtforms import (HiddenField, TextField, FieldList, IntegerField, FormField,
                     SubmitField, SelectField)
from wtforms.validators import DataRequired, Optional
from wtforms.ext.sqlalchemy.fields import QuerySelectField


class WatcherControlForm(FlaskForm):
    """ Form used to control a watcher. """
    reload = SubmitField('reload')
    start = SubmitField('start')
    stop = SubmitField('stop')
    watcher_id = HiddenField('watcher_id', validators=[DataRequired()])

    @classmethod
    def for_watcher(cls, watcher_id):
        """ Prepare form for a specific watcher """
        cls.watcher_id = HiddenField('watcher_id',
                                     validators=[DataRequired()],
                                     default=watcher_id)

        return cls


class WatcherCreateForm(FlaskForm):
    """ Form used to select which thread_class the newly created watcher will
    use
    """
    watcher_class = SelectField('Watcher Class', validators=[DataRequired()],
                                choices=[
        ('lamon.watcher.plugin.source_engine.SourceEngineWatcher',
         'Source Engine Watcher'),
        ('lamon.watcher.plugin.quake3.Quake3Watcher',
         'Quake3 Watcher'),
        ('tests.test_watcher.fake.FakeWatcher',
         'Fake Watcher')
    ])


class WatcherEditForm(FlaskForm):
    """ Form used to edit a watcher instance """
    game = QuerySelectField(allow_blank=True)

    @classmethod
    def load_config(cls, watcher_class, watcher_model):
        """ Prepare form for a specific watcher """
        cls.watcher = watcher_class
        cls.game = QuerySelectField(
            allow_blank=True, default=watcher_model.game)

        for key, opts in watcher_class.config_keys.items():
            value = None

            for i in watcher_model.config:
                if i.key == key:
                    value = i.value

            if watcher_class.config_keys[key]['type'] is int:
                field = IntegerField
            else:
                field = TextField

            if watcher_class.config_keys[key]['required']:
                validators = [DataRequired()]
            else:
                validators = [Optional()]

            setattr(cls, key, field(key, default=value, validators=validators))

        return cls



from flask import Blueprint, render_template

main_blueprint = Blueprint('main', __name__, template_folder='templates')

@main_blueprint.route('/')
def index():
    return render_template('main/index.html')

@main_blueprint.route('/stats')
def stats():
    return render_template('main/stats.html')



from valve.source.a2s import ServerQuerier, NoResponseError

from .. import Watcher, WatcherException
from lamon.models import Event, EventType


class SourceEngineWatcher(Watcher):
    """Watcher implementating communication to games based on Valve's source
    engine
    """

    config_keys = {'address': {'type': str, 'required': True},
                   'port': {'type': int, 'required': True},
                   'timeout': {'type': int, 'required': True},
                   'app_id': {'type': int, 'required': True,
                               'hint':
                               """this is the app ID of the client, not the
                               server. For example, for Team Fortress 2 440 has
                               to be used instead of 232250 which is the ID of
                               the server software."""}}

    def __init__(self, **kwargs):
        super().__init__(__name__, **kwargs)

    def runner(self):
        # Setup valve server querier
        addr = (self.config['address'], self.config['port'])
        connection_lost = False
        server_info = None

        with ServerQuerier(addr, self.config['timeout']) as server:
            while getattr(self, 'shutdown', True):
                try:
                    server_info = server.info()
                except NoResponseError:
                    self.add_event(
                        Event(type=EventType.WATCHER_CONNECTION_LOST))
                    self.logger.debug("Watcher connection failure")
                    connection_lost = True
                    continue

                if server_info['app_id'] != int(self.config['app_id']):
                    raise WatcherException(
                        "Wrong app id on server: {}".format(server_info['app_id']))
                self._updatePlayers(server.players()['players'])

                if connection_lost:  # Watcher reaquired connection
                    self.add_event(
                        Event(type=EventType.WATCHER_CONNECTION_REAQUIRED))

    def _updatePlayers(self, players):
        for p in players:
            if not p['name']:  # Valve doc mentions possible empty players
                continue

            self.add_score(p['name'], p['score'])



import socket
import re
import time

from .. import Watcher, WatcherException
from lamon.models import Event, EventType


class Quake3Watcher(Watcher):
    """ Watcher implementing communication to Quake 3 """

    config_keys = {'address': {'type': str, 'required': True},
                   'port': {'type': int, 'required': True},
                   'timeout': {'type': int, 'required': True},
                   'rcon_password': {'type': str, 'required': True}}

    def __init__(self, **kwargs):
        super().__init__(__name__, config_keys=config_keys, **kwargs)

        self.packet_prefix = ('\xff' * 4).encode()

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.settimeout(self.config['timeout'])

        self.connection_lost = False

        # Regexes. Save here so we don't compile them everytime
        self.user_re = re.compile('\s*(\d*)\s*(\d*)\s*(\d*)\s*([^\x12]*)')
        self.map_re = re.compile('map: (.*)')

    def runner(self):
        self.sock.connect((self.config['address'], self.config['port']))

        while getattr(self, 'shutdown', True):
            try:
                info = self.get_info()
            except WatcherException:
                if not self.connection_lost: # Prevent duplicate events
                    self.add_event(Event(type=EventType.WATCHER_CONNECTION_LOST))
                    self.connection_lost = True

                time.sleep(3)
                continue

            for key, value in info['players'].items():
                try:
                    self.add_score(key, value['frags'])
                except ValueError: pass

            if self.connection_lost:
                self.add_event(Event(type=EventType.WATCHER_CONNECTION_REAQUIRED))
                self.connection_lost = False
            time.sleep(3)

    def get_info(self):
        """ Get Server information

        :returns: Dict of server info
        :rtype: dict
        """
        result = {'players': {}}

        rcon_status = self.rcon('status')[1].replace('^7', '\x12').split('\n')

        for i in rcon_status[3:]:
            match = self.user_re.match(i)
            if match and match.group(4) is not '':
                result['players'][match.group(4)] = {
                    'frags': match.group(2),
                    'ping': match.group(3)
                }

        match = self.map_re.match(rcon_status[0])
        if match:
            result['map'] = match.group(1)

        return result

    def cmd(self, cmd):
        """ Execute a command on the server

        :type cmd: str
        :param cmd: Command to execute

        :returns: responseType, responseBody
        :rtype: tuple

        :raises: WatcherException
        """
        self.sock.send(b'\xFF\xFF\xFF\xFF' + cmd.encode())

        try:
            resp = self.sock.recv(4096)
        except (socket.timeout, socket.error) as e:
            raise WatcherException("Watcher connection failure")

        resp = resp[len(b'\xFF\xFF\xFF\xFF'):].decode().split('\n')

        responseType = resp[0]
        responseBody = '\n'.join(resp[1:])

        return responseType, responseBody

    def rcon(self, cmd):
        """ Execute a rcon command

        :type cmd: str
        :param cmd: RCON command to execute

        :returns: responseType, responseBody
        :rtype: tuple

        :raises WatcherException: Bas RCON Password
        """
        respType, respBody = self.cmd(
            'rcon "{}" {}'.format(self.config['rcon_password'], cmd))

        if respBody == 'Bad rconpassword.\n':
            raise WatcherException("Bad RCON Password")

        return respType, respBody
